#lang typed/racket/base

(require racket/match)

(provide (all-defined-out))

(struct Term () #:transparent)
(struct Var Term ((name : Symbol) (internal : Symbol)) #:transparent)
(struct Abs Term ((var : Symbol) (internal : Symbol) (body : Term)) #:transparent)
(struct App Term ((fun : Term) (arg : Term)) #:transparent)

(: var (-> Symbol Var))
(define (var n) (Var n n))

(: abs (-> Symbol Term Abs))
(define (abs v b) (Abs v v b))

(: app (-> Term Term App))
(define (app f a) (App f a))

(: lc-fmt (-> Term String))
(define (lc-fmt term)
  (match term
    ((Var _ n)
     (format "~a" n))
    ((Abs _ v b)
     (format "(λ ~a. ~a)" v (lc-fmt b)))
    ((App f a)
     (format "(~a ~a)" (lc-fmt f) (lc-fmt a)))))

(: lc-pp (-> Term Void))
(define (lc-pp term)
  (displayln (lc-fmt term)))
