#lang racket/base

(require parser-tools/lex
         (prefix-in : parser-tools/lex-sre)
         parser-tools/yacc
         "lc-ast.rkt")

(provide lc-parser)

(define-empty-tokens lc-syntax
  (Llam Ldot Lopar Lcpar Leof))

(define-tokens lc-var
  (Lvar))

(define-lex-abbrev latin_
  (:or (char-range "a" "z")
       (char-range "A" "Z")
       "_"))

(define-lex-abbrev latin_num
  (:: latin_
      (:* (:or latin_ numeric))))

(define lc-lex
  (lexer-src-pos
   ((eof)      'Leof)
   (whitespace (return-without-pos (lc-lex input-port)))
   ("λ"        (token-Llam))
   ("."        (token-Ldot))
   ("("        (token-Lopar))
   (")"        (token-Lcpar))
   (latin_num  (token-Lvar (string->symbol lexeme)))
   (any-char
    (error 'Lexer
           "Invalid char '~a' on line ~a col ~a."
           lexeme
           (position-line start-pos)
           (position-col start-pos)))))

(define lc-parse
  (parser
   (src-pos)
   (tokens lc-syntax lc-var)
   (start expr)
   (end Leof)
   (grammar
    (expr
     ((Lvar)
      (var $1))
     ((Llam Lvar Ldot follow-expr)
      (abs $2 $4))
     ((expr follow-expr)
      (app $1 $2))
     ((Lopar expr Lcpar) $2))
    (follow-expr
     ((Lvar)
      (var $1))
     ((Lopar expr Lcpar) $2)))
   (error
    (lambda (ok? name value sp ep)
      (error 'Parser
             "~a on line ~a col ~a."
             (if ok?
                 "Syntax error"
                 (format "Invalid token ~a~a"
                         name
                         (if value (format " (~a)" value) "")))
             (position-line sp)
             (position-col sp))))))

(define (lc-parser str)
  (let ((in (open-input-string str)))
    (port-count-lines! in)
    (lc-parse (lambda () (lc-lex in)))))
